/*: -----------------------------------------------------------------------
    ** This file was stripped of all comments and exported by Snowmobile **
    ----------------------------------------------------------------------- :*/

/*-
__intro.sql__
__authored-by: Some Chap or Lass
__authored-on: Some Day or Year
__p*_***:
**Impetus**: *SQL is older than time and isn't going anywhere; might we allow a simple markup syntax?*
-*/

/*-create table~sample_table; DDL-*/
create or replace table sample_table (
	col1 number(18,0),
	col2 number(18,0)
);
