# import link
from . import link

link_name = "SnowflakeConnection.execute_stream()"
user_text = "SnowflakeConnection.execute_stream()"
url = "https://docs.snowflake.com/en/user-guide/python-connector-example.html#using-execute-stream-to-execute-sql-scripts"

link.xref_links.update({link_name: (user_text, url)})
