# import link
from . import link

link_name = "file format"
user_text = "file format"
url = "https://docs.snowflake.com/en/user-guide/data-load-internal-tutorial-create-file-format.html"

link.xref_links.update({link_name: (user_text, url)})
