# import link
from . import link

link_name = "DataFrame"
user_text = "DataFrame"
url = "https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html"

link.xref_links.update({link_name: (user_text, url)})
