# import link
from . import link

link_name = "copy into table"
user_text = "copy into table"
url = "https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html"

link.xref_links.update({link_name: (user_text, url)})
