#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['snowmobile',
 'snowmobile.core',
 'snowmobile.core.cfg',
 'snowmobile.core.pkg_data',
 'snowmobile.core.utils']

package_data = \
{'': ['*'],
 'snowmobile': ['snowmobile.egg-info/*'],
 'snowmobile.core': ['pkg_data/.snowmobile/*', 'pkg_data/.snowmobile/DDL/*']}

install_requires = \
['snowflake-connector-python==2.3.7',
 'pandas==1.1.5',
 'pydantic==1.7.2',
 'appdirs==1.4.4',
 'toml==0.10.1',
 'tabulate==0.8.7',
 'brotli==1.0.9',
 'sqlparse==0.3.1',
 'ipython==7.19.0',
 'ipykernel==5.4.2']

extras_require = \
{'docs': ['appdirs==1.4.4',
          'argon2-cffi==20.1.0',
          'asn1crypto==1.4.0',
          'astroid==2.4.2',
          'async-generator==1.10',
          'atomicwrites==1.4.0',
          'auditwheel==3.1.1',
          'azure-common==1.1.25',
          'azure-core==1.8.2',
          'azure-storage-blob==12.5.0',
          'azure-storage-common==2.1.0',
          'Babel==2.9.0',
          'backcall==0.2.0',
          'beautifulsoup4==4.9.3',
          'bleach==3.2.1',
          'boto3==1.15.18',
          'botocore==1.18.18',
          'Brotli==1.0.9',
          'CacheControl==0.12.6',
          'cachy==0.3.0',
          'certifi==2020.12.5',
          'cffi==1.14.5',
          'chardet==3.0.4',
          'cleo==0.7.6',
          'clikit==0.4.3',
          'colorama==0.4.4',
          'coverage==5.5',
          'cryptography==2.9.2',
          'css-html-js-minify==2.5.5',
          'decorator==4.4.2',
          'defusedxml==0.6.0',
          'docutils==0.15.2',
          'entrypoints==0.3',
          'flit==3.1.0',
          'flit-core==3.1.0',
          'gitdb==4.0.5',
          'GitPython==3.1.12',
          'html5lib==1.0.1',
          'idna==2.10',
          'ijson==2.6.1',
          'imagesize==1.2.0',
          'importlib-metadata==1.1.3',
          'iniconfig==1.1.1',
          'ipykernel==5.4.2',
          'ipython==7.19.0',
          'ipython-genutils==0.2.0',
          'ipywidgets==7.6.3',
          'isodate==0.6.0',
          'jedi==0.18.0',
          'Jinja2==2.11.2',
          'jmespath==0.10.0',
          'jsonschema==3.2.0',
          'jupyter-cache==0.4.2',
          'jupyter-client==6.1.11',
          'jupyter-core==4.7.0',
          'jupyter-sphinx==0.3.1',
          'jupyterlab-widgets==1.0.0',
          'keyring==20.0.1',
          'lazy-object-proxy==1.4.3',
          'lockfile==0.12.2',
          'lxml==4.6.2',
          'markdown-it-py==0.6.2',
          'MarkupSafe==1.1.1',
          'mdit-py-plugins==0.2.6',
          'mistune==0.8.4',
          'msgpack==1.0.0',
          'msrest==0.6.19',
          'myst-nb==0.12.0',
          'myst-parser==0.13.5',
          'nbclient==0.5.1',
          'nbconvert==5.6.1',
          'nbdime==2.1.0',
          'nbformat==5.1.2',
          'nest-asyncio==1.4.3',
          'notebook==6.2.0',
          'numpy==1.18.4',
          'oauthlib==3.1.0',
          'oscrypto==1.2.1',
          'packaging==20.8',
          'pandas==1.1.5',
          'pandocfilters==1.4.3',
          'parso==0.8.1',
          'pastel==0.2.0',
          'pexpect==4.8.0',
          'pickleshare==0.7.5',
          'pkginfo==1.5.0.1',
          'pluggy==0.13.1',
          'pockets==0.9.1',
          'prometheus-client==0.9.0',
          'prompt-toolkit==3.0.10',
          'ptyprocess==0.6.0',
          'py==1.10.0',
          'pyarrow==0.17.0',
          'pycparser==2.20',
          'pycryptodomex==3.9.8',
          'pydantic==1.7.2',
          'pyelftools==0.26',
          'Pygments==2.7.4',
          'PyJWT==1.7.1',
          'pylev==1.3.0',
          'pyOpenSSL==19.1.0',
          'pyparsing==2.4.7',
          'pyrsistent==0.14.11',
          'pytest==6.2.2',
          'python-dateutil==2.8.1',
          'python-slugify==4.0.1',
          'pytz==2020.1',
          'pywin32-ctypes==0.2.0',
          'pywinpty==0.5.7',
          'PyYAML==5.4',
          'pyzmq==21.0.1',
          'requests<2.24.0',
          'requests-oauthlib==1.3.0',
          'requests-toolbelt==0.8.0',
          's3transfer==0.3.3',
          'Send2Trash==1.5.0',
          'shellingham==1.3.2',
          'six==1.15.0',
          'smmap==3.0.4',
          'snowballstemmer==2.0.0',
          'snowflake-connector-python==2.3.7',
          'soupsieve==2.1',
          'Sphinx==3.5.3',
          'sphinx-autoapi==1.7.0',
          'sphinx-copybutton==0.3.1',
          'sphinx-material==0.0.32',
          'sphinx-panels==0.5.2',
          'sphinx-togglebutton==0.2.3',
          'sphinxcontrib-applehelp==1.0.2',
          'sphinxcontrib-devhelp==1.0.2',
          'sphinxcontrib-dotnetdomain==0.4',
          'sphinxcontrib-golangdomain==0.2.0.dev0',
          'sphinxcontrib-htmlhelp==1.0.3',
          'sphinxcontrib-jsmath==1.0.1',
          'sphinxcontrib-napoleon==0.7',
          'sphinxcontrib-qthelp==1.0.3',
          'sphinxcontrib-serializinghtml==1.1.4',
          'SQLAlchemy==1.3.22',
          'sqlparse==0.3.1',
          'tabulate==0.8.7',
          'terminado==0.9.2',
          'testpath==0.4.4',
          'text-unidecode==1.3',
          'toml==0.10.1',
          'tomlkit==0.5.11',
          'tornado==6.1',
          'traitlets==5.0.5',
          'typed-ast==1.4.2',
          'typing-extensions==3.7.4.3',
          'Unidecode==1.1.2',
          'urllib3==1.25.11',
          'wcwidth==0.2.5',
          'webencodings==0.5.1',
          'widgetsnbextension==3.5.1',
          'wincertstore==0.2',
          'wrapt==1.12.1',
          'zipp==3.1.0'],
 'test': ['pytest >= 5.0.0',
          'pytest-cov>=2.10.1',
          'flit>=3.0.0',
          'flit-core>=3.0.0']}

setup(name='snowmobile',
      version='0.2.0b23',
      description='An analytics-focused wrapper around the snowflake.connector for Python.',
      author='Grant Murray',
      author_email='gmurray203@gmail.com',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7.0',
     )
