# import link
from . import link

link_name = "DictCursor"
user_text = "DictCursor"
url = "https://docs.snowflake.com/en/user-guide/python-connector-api.html#cursor"

link.xref_links.update({link_name: (user_text, url)})
