-- noinspection SqlResolveForFile
-- ..docs/snippets/getting_started/any_other_table.sql


create or replace table gem7318.any_other_table (
	col1 number(18,0),
	col2 number(18,0),
	insert_tmstmp timestamp_tz
);
